-- modex.e
-- Mode X Graphics Library
-- Pete Eberlein <xseal@harborside.com>
-- 8 Dec 97

global constant allowed_widths = {256,320,360,376,400}
global constant allowed_heights = {200,224,240,256,270,300,360,400,480,564,600}

include graphics.e
include machine.e

integer dest_width, dest_height --used for clipping in pixel()

global constant screen = allocate(24)
poke4(screen, repeat(0, 6))

atom current_source, current_dest, current_clip
current_source = screen
current_dest = screen
    
constant video_mem = #A0000
constant sequ = #3C4, --sequencer register base
         crtc = #3D4, --CRT controller register base
         misc = #3C2, --miscellaneous output register
         ins1 = #3DA, --input status #1 register
         attr = #3C0  --attribute register 

function allocate_proc(sequence data)
    atom handle
    handle = allocate(length(data))
    if handle then
--        lock_memory(handle, length(data)) --is this necessary?
        poke(handle, data)
        return handle
    end if
    printf(1, "Unable to allocate memory for length-%d sequence...\nProgram aborted.\n", {length(data)})
    abort(1)
end function


constant in_byte_result = allocate(1)
constant in_byte_proc = allocate_proc(
   {#50,                    --push eax
    #52,                    --push edx
    #BA,#00,#00,#00,#00,    --mov edx, port (3)
    #EC,                    --in al, dx
    #A2}                    --mov result, al
    & int_to_bytes(in_byte_result) & {
    #5A,                    --pop edx
    #58,                    --pop eax
    #C3})                   --ret
constant out_byte_proc = allocate_proc(
   {#50,                    --push eax
    #52,                    --push edx
    #BA,#00,#00,#00,#00,    --mov edx, port (3)
    #B0,#00,                --mov al, data (8)
    #EE,                    --out dx, al
    #5A,                    --pop edx
    #58,                    --pop eax
    #C3})                   --ret
constant out_word_proc = allocate_proc(
   {#50,                    --push eax
    #52,                    --push edx
    #BA,#00,#00,#00,#00,    --mov edx, port (3)
    #B8,#00,#00,#00,#00,    --mov eax, data (8)
    #66,#EF,                --out dx, ax
    #5A,                    --pop edx
    #58,                    --pop eax
    #C3})                   --ret

function in_byte(integer port)
    poke4(in_byte_proc + 3, port)
    return peek(in_byte_result)
end function

procedure out_byte(integer port, object data)
    poke4(out_byte_proc + 3, port)
    if atom(data) then
        data = {data}
    end if
    for i = 1 to length(data) do
        poke(out_byte_proc + 8, data[i])
        call(out_byte_proc)
    end for
end procedure

procedure out_word(integer port, object data)
    poke4(out_word_proc + 3, port)
    if atom(data) then
        data = {data}
    end if
    for i = 1 to length(data) do
        poke4(out_word_proc + 8, data[i])
        call(out_word_proc)
    end for
end procedure

-- returns machine codes for "ebx = ebx + edx * width"
function fast_asm_mul(integer width)
    integer count  sequence s
    count = 0      s = {}
    while width do
        if and_bits(width, 1) then
            s = s & {#C1,#E2,count, --shl edx, count
                     #01,#D3}       --add ebx, edx
            count = 0
        end if
        count = count + 1
        width = floor(width / 2)
    end while
    return s
end function

-----------------------------------------------------------------
--SCREENS--------------------------------------------------------
-----------------------------------------------------------------

constant --offsets for data in a screen handle
    screen_width = 0,
    screen_height = 4,
    screen_source = 8,
    screen_dest = 12

-- creates a virtual screen and returns its handle
global function virtual_screen(integer width, integer height)
    atom screen_mem, handle
    sequence source_code

    width = floor((width+3)/4)
    source_code = 
        fast_asm_mul(width*4)&
       {#88,#DA,                --mov dl, bl
        #83,#E2,#03,            --and edx, 3
        #C1,#E2,#02,            --shl edx, 2
        #C1,#FB,#02,            --sar ebx, 2
        #03,#9A,#00,#00,#00,#00,--add ebx, dword ptr [video_mem + edx]
        #C3}                    --ret
    handle = allocate(32 + length(source_code) + 4*width*height)
    screen_mem = handle + 32 + length(source_code)
    poke4(handle + screen_width, width*4)
    poke4(handle + screen_height, height)
    poke4(handle + screen_source, handle + 32)
    poke4(handle + screen_dest, handle + 32)
    poke4(handle + 16, screen_mem + height*width*{0,1,2,3})
    poke(handle + 32, source_code)
    poke4(screen_mem - 5,  handle + 16)
    mem_set(screen_mem, 0, width*height)

    return handle
end function

-----------------------------------------------------------------
--MODES----------------------------------------------------------
-----------------------------------------------------------------

constant
    screen_pan_to = 16,
    screen_originate = 20

global procedure set_width(integer width)
    atom source_code, dest_code, pan_to_code, originate_code
    out_byte(crtc, #13)
    width = floor((width + 7) / 8)
    out_byte(crtc+1, width)
    poke(#44A, remainder(width, 256) & floor(width / 256)) -- BIOS Video Columns
    width = width * 2
    -- initialize the screen handle
    for i = screen+screen_source to screen+screen_originate by 4 do
        if peek4u(i) then
            free(peek4u(i))
        end if
    end for
    source_code = allocate_proc(
              {#50,                    --push eax
               #81,#C3,#00,#00,#00,#00,--add ebx, dword 0
               #88,#DC,                --mov ah, bl
               #80,#E4,#03,            --and ah, 3
               #C1,#FB,#02,            --sar ebx, 2
               #81,#C3}&int_to_bytes(video_mem)&--add ebx, dword video_mem (34)
               fast_asm_mul(width)&{   --ebx = ebx + edx * width
               #BA,#CE,#03,#00,#00,    --mov edx, #3CE
               #B0,#04,                --mov al, 4
               #66,#EF,                --out dx, ax
               #58,                    --pop eax
               #C3})                   --ret
    dest_code = allocate_proc(
              {#50,                    --push eax
               #51,                    --push ecx
               #81,#C3,#00,#00,#00,#00,--add ebx, dword 0
               #88,#D9,                --mov cl, bl
               #80,#E1,#03,            --and cl, 3
               #C1,#FB,#02,            --sar ebx, 2
               #81,#C3}&int_to_bytes(video_mem)&--add ebx, dword video_mem
               fast_asm_mul(width)&{   --ebx = ebx + edx * width
               #BA,#C4,#03,#00,#00,    --mov edx, #3C4
               #B4,#01,                --mov ah, 1
               #B0,#00,                --mov al, 0
               #D3,#E0,                --shl eax, cl
               #B0,#02,                --mov al, 2
               #66,#EF,                --out dx, ax
               #59,                    --pop ecx
               #58,                    --pop eax
               #C3})                   --ret
    pan_to_code = allocate_proc(
       {#81,#C3,#00,#00,#00,#00}&--add ebx, dword 0
        fast_asm_mul(width*4)&    --bx = bx + dx * width*4
       {#BA,#DA,#03,#00,#00,    --mov edx, #3DA
        #EC,                    --in al, dx
        #B2,#C0,                --mov dl, #C0
        #B0,#33,                --mov al, #33
        #EE,                    --out dx, al
        #88,#D8,                --mov al, bl
        #24,#03,                --and al, 3
        #D0,#E0,                --shl al, 1
        #EE,                    --out dx, al
        #C1,#EB,#02,            --shr ebx, 2
        #B2,#D4,                --mov dl, #D4
        #B0,#0C,                --mov al, #0C
        #88,#FC,                --mov ah, bh
        #66,#EF,                --out dx, ax
        #B0,#0D,                --mov al, #0D
        #88,#DC,                --mov ah, bl
        #66,#EF,                --out dx, ax
        #C3})                   --ret
        
    originate_code = allocate_proc(
        fast_asm_mul(width*4)&  --bx = bx + dx * width * 4
       {#89,#D8,                --mov eax, ebx
        #A3}&int_to_bytes(source_code+3)&--mov [source_add], eax (3)
       {#A3}&int_to_bytes(dest_code+4)&--mov [dest_add], eax (8)
       {#A3}&int_to_bytes(pan_to_code+2)&--mov [pan_add], eax (13)
       {#C3})                   --ret

    poke4(screen + screen_width, width*4)
    poke4(screen + screen_height, floor(65536 / width))
    poke4(screen + screen_source, source_code)
    poke4(screen + screen_dest, dest_code)
    poke4(screen + screen_pan_to, pan_to_code)
    poke4(screen + screen_originate, originate_code)
end procedure

global function mode_x(integer width, integer height)
    sequence h_regs, v_regs
    integer misc_value
    misc_value = #23
    if width = 256 then
        h_regs = {#5F00,#3F01,#4002,#8203,#4A04,#9A05,#2013}
        misc_value = misc_value + #00
    elsif width = 320 then
        h_regs = {#5F00,#4F01,#5002,#8203,#5404,#8005,#2813}
        misc_value = misc_value + #00
    elsif width = 360 then
        h_regs = {#6B00,#5901,#5A02,#8E03,#5E04,#8A05,#2D13}
        misc_value = misc_value + #04
    elsif width = 376 then
        h_regs = {#6E00,#5D01,#5E02,#9103,#6204,#8F05,#2F13}
        misc_value = misc_value + #04
    elsif width = 400 then
        h_regs = {#7100,#6301,#6402,#9203,#6504,#8205,#3213}
        misc_value = misc_value + #04
    else
        return 1  --unsupported width
    end if
    if height = 200 then
        v_regs = {#BF06,#1F07,#4109,#9C10,#8E11,#8F12,#9615,#B916}
        misc_value = misc_value + #40
    elsif height = 224 then
        v_regs = {#0B06,#3E07,#4109,#DA10,#9C11,#BF12,#C715,#0416}
        misc_value = misc_value + #C0
    elsif height = 240 then
        v_regs = {#0D06,#3E07,#4109,#EA10,#AC11,#DF12,#E715,#0616}
        misc_value = misc_value + #C0
    elsif height = 256 then
        v_regs = {#2306,#B207,#6109,#0A10,#AC11,#FF12,#0715,#1A16}
        misc_value = misc_value + #C0
    elsif height = 270 then
        v_regs = {#3006,#F007,#6109,#2010,#A911,#1B12,#1F15,#2F16}
        misc_value = misc_value + #C0
--    elsif height = 282 then
--        v_regs = {#6206,#0F07,#6109,#3710,#8911,#3312,#3C15,#5C16}
--        misc_value = misc_value + #C0
    elsif height = 300 then
        v_regs = {#4606,#1F07,#4009,#3110,#8011,#2B12,#2F15,#4416}
        misc_value = misc_value + #40
--    elsif height = 308 then
--        v_regs = {#6206,#0F07,#4009,#3710,#8911,#3312,#3C15,#5C16}
--        misc_value = misc_value + #C0
    elsif height = 360 then
        v_regs = {#BF06,#1F07,#4009,#8810,#8511,#6712,#6D15,#BA16}
        misc_value = misc_value + #40
    elsif height = 400 then
        v_regs = {#BF06,#1F07,#4009,#9C10,#8E11,#8F12,#9615,#B916}
        misc_value = misc_value + #40
    elsif height = 480 then
        v_regs = {#0D06,#3E07,#4009,#EA10,#AC11,#DF12,#E715,#0616}
        misc_value = misc_value + #C0
    elsif height = 564 then
        v_regs = {#6206,#F007,#6009,#3710,#8911,#3312,#3C15,#5C16}
        misc_value = misc_value + #C0
    elsif height = 600 then
        v_regs = {#7006,#F007,#6009,#5B10,#8C11,#5712,#5815,#7016}
        misc_value = misc_value + #C0
    else 
        return 1  --unsupported height
    end if

    if graphics_mode(19) = 0 then
        
        -- set planar mode
        out_word(sequ, #0604)

        -- clear the screen
        out_word(sequ, #0F02)
        poke(video_mem, repeat(0, #10000))

        -- more to set the planar mode
        out_word(sequ, #0100)
        out_byte(misc, misc_value)
        out_word(sequ, #0300)

        -- allow access to CRTC protected registers
        out_byte(crtc, #11)
        out_byte(crtc+1, and_bits(in_byte(crtc+1), #7F))

        -- set horizontal and vertical registers
        out_word(crtc, h_regs & v_regs & {#0008,#0014,#E317})

        -- set the overscan color to 0
        if in_byte(ins1) then end if
        out_byte(attr, {#31, 0})

        set_width(width)
        return 0
    else
        return 1
    end if
end function


constant 
    get_addr_result = allocate(4),
    get_addr_proc = allocate_proc(
       {#60,                    --pusha
        #BB,#00,#00,#00,#00,    --mov ebx, x (2)
        #BA,#00,#00,#00,#00,    --mov edx, y (7)
        #B8,#00,#00,#00,#00,    --mov eax, handle (12)
        #FF,#50,#00,            --call near dword ptr [eax+handleofs] (18)
        #89,#1D}&int_to_bytes(get_addr_result)&{--mov [result], ebx (21)
        #61,                    --popa
        #C3})                   --ret

function get_source_addr(sequence pos)
    poke4(get_addr_proc + 2, pos[1])
    poke4(get_addr_proc + 7, pos[2])
    poke4(get_addr_proc + 12, current_source)
    poke(get_addr_proc + 18, screen_source)
    call(get_addr_proc)
    return peek4u(get_addr_result)
end function

function get_dest_addr(sequence pos)
    poke4(get_addr_proc + 2, pos[1])
    poke4(get_addr_proc + 7, pos[2])
    poke4(get_addr_proc + 12, current_dest)
    poke(get_addr_proc + 18, screen_dest)
    call(get_addr_proc)
    return peek4u(get_addr_result)
end function
    
-----------------------------------------------------------------
--COMMANDS-------------------------------------------------------
-----------------------------------------------------------------

global constant
    copy = 1,
    clearcopy = 2,
    destcopy = 3,
    mixcopy = 4,
    clearmixcopy = 5,
    pan_to = 6,
    originate = 7,
    wait_retrace = 10


-- offsets for the command structure
constant 
    start_offset = 11,
    loop_offset = 12, 
    prev_command = 3,
    next_command = 13,
    z_order = 7

constant 
    loop_command = allocate(4),
    return_proc = allocate_proc(
       {#C3,                    --ret
        #90,                    --nop
        #90,                    --nop
        #00,#00,#00,#00,        --prev_command: dd ? (3)
        #FF,#FF,#FF,#7F,        --z_order: dd #7FFFFFFF (7)
        #60,                    --start_offset: pusha 
        #61,                    --loop_offset: popa
        #C3})                   --ret

poke4(loop_command, return_proc + loop_offset)

constant command_header =
   {#60,                    --pusha
    #EB,#16,                --jmp short label1
    #00,#00,#00,#00,        --prev_command: dd ? (3)
    #00,#00,#00,#00,        --z_order: dd ? (7)
    #60,                    --start_offset: pusha 
    #B8,#00,#00,#00,#00,    --loop_offset: mov eax, next_command (13)
    #83,#C0,loop_offset,    --add eax, loop_offset
    #A3}&int_to_bytes(loop_command)--mov [loop_command], eax
                            --label1:
constant command_footer =
   {#FF,#25}&int_to_bytes(loop_command)--jmp near dword ptr [loop_command]

constant generic_copy_code =
   {#BE,#00,#00,#00,#00,    --mov esi, source_screen (1)
    #BF,#00,#00,#00,#00,    --mov edi, dest_screen (6)
    #B8,#00,#00,#00,#00,    --mov eax, command_handle (11)
    #BB,#00,#00,#00,#00,    --mov ebx, clip_proc (16)
    #FF,#D3,                --call near ebx
    #BB,#00,#00,#00,#00,    --mov ebx, copy_proc (23)
    #FF,#D3}                --call near ebx

constant 
    generic_source_screen  = length(command_header) + 1,
    generic_dest_screen    = length(command_header) + 6,
    generic_command_handle = length(command_header) + 11,
    generic_clip_proc      = length(command_header) + 16,
    generic_copy_proc      = length(command_header) + 23,
    generic_data = length(command_header & generic_copy_code & command_footer)

constant 
    source_x = generic_data,
    source_y = generic_data + 4,
    source_w = generic_data + 8,
    source_h = generic_data + 12,
    dest_x = generic_data + 16,
    dest_y = generic_data + 20,
    source_mixtable = generic_data + 24,
    generic_length = generic_data + 24,
    generic_mixtable_length = generic_data + 28


constant wait_retrace_code = 
   {#BA,#DA,3,0,0,--mov edx,#3DA
    #EC,          --label1: in al,dx
    #24,#08,      --and al,8
    #75,#FB,      --jnz label1
    #EC,          --label2: in al,dx
    #24,#08,      --and al,8
    #74,#FB}      --jz label2


global    
constant                      -- source | dest
    clip_data = allocate(24), -- x,y,w,h, x,y
    clip_proc = allocate_proc(
   {#8B,#58,source_x,       --mov ebx, [eax+byte source_x] (2)
    #8B,#48,source_w,       --mov ecx, [eax+byte source_w] (5)
    #8B,#50,dest_x,         --mov edx, [eax+byte dest_x] (8)
    #83,#FA,#00,            --cmp edx, 0
    #7D,#06,                --jge label1
    #01,#D1,                --add ecx, edx
    #29,#D3,                --sub ebx, edx
    #31,#D2,                --xor edx, edx
    #8B,#2F,                --label1: mov ebp, [edi]
    #29,#D5,                --sub ebp, edx
    #29,#CD,                --sub ebp, ecx
    #83,#FD,#00,            --cmp ebp, 0
    #7D,#02,                --jge label2
    #01,#E9,                --add ecx, ebp
    #89,#1D}&int_to_bytes(clip_data)&{--label2: mov [clip_x], ebx (35)
    #89,#0D}&int_to_bytes(clip_data+8)&{--mov [clip_w], ecx (41)
    #89,#15}&int_to_bytes(clip_data+16)&{--mov [clip_dx], edx (47)
    #83,#F9,#00,            --cmp ecx, 0
    #7F,#07,                --jg label2b
    #89,#0D}&int_to_bytes(clip_data+12)&{--mov [clip_h], ecx (58)
    #C3,                    --ret
    #8B,#58,source_y,       --label2b: mov ebx, [eax+byte source_y] (65)
    #8B,#48,source_h,       --mov ecx, [eax+byte source_h] (68)
    #8B,#50,dest_y,         --mov edx, [eax+byte dest_y] (71)
    #83,#FA,#00,            --cmp edx, 0
    #7D,#06,                --jge label3
    #01,#D1,                --add ecx, edx
    #29,#D3,                --sub ebx, edx
    #31,#D2,                --xor edx, edx
    #8B,#6F,#04,            --label3: mov ebp, [edi+4]
    #29,#D5,                --sub ebp, edx
    #29,#CD,                --sub ebp, ecx
    #83,#FD,#00,            --cmp ebp, 0
    #7D,#02,                --jge label4
    #01,#E9,                --add ecx, ebp
    #89,#1D}&int_to_bytes(clip_data+4)&{--label4: mov [clip_y], ebx (99)
    #89,#0D}&int_to_bytes(clip_data+12)&{--mov [clip_h], ecx (105)
    #89,#15}&int_to_bytes(clip_data+20)&{--mov [clip_dy], edx (111)
    #C3})                   --ret
constant noclip_proc = allocate_proc(
   {#56,                    --push esi
    #57,                    --push edi
    #89,#C6,                --mov esi, eax
    #83,#C6,source_x,       --add esi, byte source_x (4)
    #BF}&int_to_bytes(clip_data)&{--mov edi, clip_data (6)
    #B9,#06,#00,#00,#00,    --mov ecx, 6
    #F3,#A5,                --rep movsd
    #5F,                    --pop edi
    #5E,                    --pop esi
    #C3})                   --ret
        
constant copy_proc = allocate_proc(
   {#B8}&int_to_bytes(clip_data)&{--mov eax, clip_data (1)
    #8B,#58,#0C,            --mov ebx, [eax+12]
    #83,#FB,#00,            --cmp ebx, 0
    #7D,#01,                --jge label0
    #C3,                    --ret
    #89,#1D,#00,#00,#00,#00,--label0: mov [aheight], ebx (16)
    #83,#40,#08,#04,        --add dword ptr [eax+8], 4
    #B9,#04,#00,#00,#00,    --mov ecx, 4
    #51,                    --label1: push ecx
    #56,                    --push esi
    #57,                    --push edi
    #50,                    --push eax
    #89,#C1,                --mov ecx, eax
    #FF,#49,#08,            --dec dword ptr [ecx+8]
    #8B,#59,#08,            --mov ebx, [ecx+8]
    #C1,#EB,#02,            --shr ebx, 2
    #8B,#06,                --mov eax, [esi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [asource_add], eax (52)
    #8B,#07,                --mov eax, [edi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [adest_add], eax (64)
    #8B,#19,                --mov ebx, [ecx]
    #FF,#01,                --inc dword ptr [ecx]
    #8B,#51,#04,            --mov edx, [ecx+4]
    #FF,#56,#08,            --call near dword ptr [esi+8]
    #89,#DE,                --mov esi, ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #FF,#41,#10,            --inc dword ptr [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#0C,            --call near dword ptr [edi+12]
    #89,#DF,                --mov edi, ebx
    #8B,#41,#08,            --mov eax, [ecx+8]
    #C1,#E8,#02,            --shr eax, 2
    #F7,#DB,                --neg ebx
    #83,#E3,#03,            --and ebx, 3
    #88,#1D,#00,#00,#00,#00,--mov [aline1], bl (107)
    #29,#D8,                --sub eax, ebx
    #88,#C3,                --mov bl, al
    #C1,#E8,#02,            --shr eax, 2
    #A3,#00,#00,#00,#00,    --mov [aline2], eax (119)
    #80,#E3,#03,            --and bl, 3
    #88,#1D,#00,#00,#00,#00,--mov [aline3], bl (128)
    #B9,#00,#00,#00,#00,    --mov ecx, height (133)
    #EB,#07,                --jmp label3
    #58,                    --label2: pop eax
    #5F,                    --pop edi
    #5E,                    --pop esi
    #59,                    --pop ecx
    #E2,#8C,                --loop label1
    #C3,                    --ret
    #51,                    --label3: push ecx
    #31,#C9,                --xor ecx, ecx
    #B1,#00,                --mov cl, line1 (150)
    #F3,#A4,                --rep movsb
    #B9,#00,#00,#00,#00,    --mov ecx, line2 (154)
    #F3,#A5,                --rep movsd
    #B1,#00,                --mov cl, line3 (161)
    #F3,#A4,                --rep movsb
    #59,                    --pop ecx
    #81,#C6,#00,#00,#00,#00,--add esi, source_add (167)
    #81,#C7,#00,#00,#00,#00,--add edi, dest_add (173)
    #E2,#DF,                --loop label3
    #EB,#D6})               --jmp short label2
poke4(copy_proc + 16, copy_proc + 133) --height
poke4(copy_proc +107, copy_proc + 150) --line1
poke4(copy_proc +119, copy_proc + 154) --line2
poke4(copy_proc +128, copy_proc + 161) --line3
poke4(copy_proc + 52, copy_proc + 167) --source_add
poke4(copy_proc + 64, copy_proc + 173) --dest_add

constant clearcopy_proc = allocate_proc(
   {#B8}&int_to_bytes(clip_data)&{--mov eax, clip_data (1)
    #8B,#58,#0C,            --mov ebx, [eax+12]
    #83,#FB,#00,            --cmp ebx, 0
    #7D,#01,                --jge label0
    #C3,                    --ret
    #89,#1D,#00,#00,#00,#00,--label0: mov [aheight], ebx (16)
    #83,#40,#08,#04,        --add dword ptr [eax+8], 4
    #B9,#04,#00,#00,#00,    --mov ecx, 4
    #51,                    --label1: push ecx
    #56,                    --push esi
    #57,                    --push edi
    #50,                    --push eax
    #89,#C1,                --mov ecx, eax
    #FF,#49,#08,            --dec dword ptr [ecx+8]
    #8B,#59,#08,            --mov ebx, [ecx+8]
    #C1,#EB,#02,            --shr ebx, 2
    #89,#1D,#00,#00,#00,#00,--mov [aline], ebx (46)
    #8B,#06,                --mov eax, [esi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [asource_add], eax (58)
    #8B,#07,                --mov eax, [edi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [adest_add], eax (70)
    #8B,#19,                --mov ebx, [ecx]
    #FF,#01,                --inc dword ptr [ecx]
    #8B,#51,#04,            --mov edx, [ecx+4]
    #FF,#56,#08,            --call near dword ptr [esi+8]
    #89,#DE,                --mov esi, ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #FF,#41,#10,            --inc dword ptr [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#0C,            --call near dword ptr [edi+12]
    #89,#DF,                --mov edi, ebx
    #B9,#00,#00,#00,#00,    --mov ecx, height (101)
    #EB,#07,                --jmp label3
    #58,                    --label2: pop eax
    #5F,                    --pop edi
    #5E,                    --pop esi
    #59,                    --pop ecx
    #E2,#AC,                --loop label1
    #C3,                    --ret
    #51,                    --label3: push ecx
    #B9,#00,#00,#00,#00,    --mov ecx, line (116)
    #AC,                    --label4: lodsb
    #20,#C0,                --and al,al
    #74,#05,                --jz label5
    #AA,                    --stosb
    #E2,#F8,                --loop label4
    #EB,#03,                --jmp label6
    #47,                    --label5: inc edi
    #E2,#F3,                --loop label4
    #59,                    --label6: pop ecx
    #81,#C6,#00,#00,#00,#00,--add esi, source_add (136)
    #81,#C7,#00,#00,#00,#00,--add edi, dest_add (142)
    #E2,#DE,                --loop label3
    #EB,#D5})               --jmp short label2
poke4(clearcopy_proc + 16, clearcopy_proc + 101) --height
poke4(clearcopy_proc + 46, clearcopy_proc + 116) --line
poke4(clearcopy_proc + 58, clearcopy_proc + 136) --source_add
poke4(clearcopy_proc + 70, clearcopy_proc + 142) --dest_add

constant destcopy_proc = allocate_proc(
   {#B8}&int_to_bytes(clip_data)&{--mov eax, clip_data (1)
    #8B,#58,#0C,            --mov ebx, [eax+12]
    #83,#FB,#00,            --cmp ebx, 0
    #7D,#01,                --jge label0
    #C3,                    --ret
    #89,#1D,#00,#00,#00,#00,--label0: mov [aheight], ebx (16)
    #83,#40,#08,#04,        --add dword ptr [eax+8], 4
    #B9,#04,#00,#00,#00,    --mov ecx, 4
    #51,                    --label1: push ecx
    #56,                    --push esi
    #57,                    --push edi
    #50,                    --push eax
    #89,#C1,                --mov ecx, eax
    #FF,#49,#08,            --dec dword ptr [ecx+8]
    #8B,#59,#08,            --mov ebx, [ecx+8]
    #C1,#EB,#02,            --shr ebx, 2
    #89,#1D,#00,#00,#00,#00,--mov [aline], ebx (46)
    #8B,#06,                --mov eax, [esi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [asource_add], eax (58)
    #8B,#07,                --mov eax, [edi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [adest_add], eax (70)
    #8B,#19,                --mov ebx, [ecx]
    #FF,#01,                --inc dword ptr [ecx]
    #8B,#51,#04,            --mov edx, [ecx+4]
    #FF,#56,#08,            --call near dword ptr [esi+8]
    #89,#DE,                --mov esi, ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#0C,            --call near dword ptr [edi+12]
    #53,                    --push ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #FF,#41,#10,            --inc dword ptr [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#08,            --call near dword ptr [edi+8]
    #5F,                    --pop edi
    #B9,#00,#00,#00,#00,    --mov ecx, height (110)
    #EB,#07,                --jmp label3
    #58,                    --label2: pop eax
    #5F,                    --pop edi
    #5E,                    --pop esi
    #59,                    --pop ecx
    #E2,#A3,                --loop label1
    #C3,                    --ret
    #51,                    --label3: push ecx
    #B9,#00,#00,#00,#00,    --mov ecx, line (125)
    #8A,#03,                --label4: mov al, [ebx]
    #43,                    --inc ebx
    #20,#C0,                --and al,al
    #75,#05,                --jnz label5
    #A4,                    --movsb
    #E2,#F6,                --loop label4
    #EB,#04,                --jmp label6
    #47,                    --label5: inc edi
    #46,                    --inc esi
    #E2,#F0,                --loop label4
    #59,                    --label6: pop ecx
    #81,#C6,#00,#00,#00,#00,--add esi, source_add (148)
    #B8,#00,#00,#00,#00,    --mov eax, dest_add (153)
    #01,#C7,                --add edi, eax
    #01,#C3,                --add ebx, eax
    #E2,#D8,                --loop label3
    #EB,#CF})               --jmp short label2
poke4(destcopy_proc + 16, destcopy_proc + 110) --height
poke4(destcopy_proc + 46, destcopy_proc + 125) --line
poke4(destcopy_proc + 58, destcopy_proc + 148) --source_add
poke4(destcopy_proc + 70, destcopy_proc + 153) --dest_add

constant mixcopy_proc = allocate_proc(
   {#8B,#58,source_mixtable,            --mov ebx, [eax+source_mixtable]
    #89,#1D,#00,#00,#00,#00,--mov [alookup], ebx (5)
    #B8}&int_to_bytes(clip_data)&{--mov eax, clip_data (1)
    #8B,#58,#0C,            --mov ebx, [eax+12]
    #83,#FB,#00,            --cmp ebx, 0
    #7D,#01,                --jge label0
    #C3,                    --ret
    #89,#1D,#00,#00,#00,#00,--label0: mov [aheight], ebx (25)
    #83,#40,#08,#04,        --add dword ptr [eax+8], 4
    #B9,#04,#00,#00,#00,    --mov ecx, 4
    #51,                    --label1: push ecx
    #56,                    --push esi
    #57,                    --push edi
    #50,                    --push eax
    #89,#C1,                --mov ecx, eax
    #FF,#49,#08,            --dec dword ptr [ecx+8]
    #8B,#59,#08,            --mov ebx, [ecx+8]
    #C1,#EB,#02,            --shr ebx, 2
    #89,#1D,#00,#00,#00,#00,--mov [aline], ebx (55)
    #8B,#06,                --mov eax, [esi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [asource_add], eax (67)
    #8B,#07,                --mov eax, [edi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [adest_add], eax (79)
    #8B,#19,                --mov ebx, [ecx]
    #FF,#01,                --inc dword ptr [ecx]
    #8B,#51,#04,            --mov edx, [ecx+4]
    #FF,#56,#08,            --call near dword ptr [esi+8]
    #89,#DE,                --mov esi, ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#0C,            --call near dword ptr [edi+12]
    #53,                    --push ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #FF,#41,#10,            --inc dword ptr [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#08,            --call near dword ptr [edi+8]
    #5F,                    --pop edi
    #B9,#00,#00,#00,#00,    --mov ecx, height (119)
    #31,#C0,                --xor eax, eax
    #EB,#07,                --jmp label3
    #58,                    --label2: pop eax
    #5F,                    --pop edi
    #5E,                    --pop esi
    #59,                    --pop ecx
    #E2,#A1,                --loop label1
    #C3,                    --ret
    #51,                    --label3: push ecx
    #B9,#00,#00,#00,#00,    --mov ecx, line (136)
    #AC,                    --label4: lodsb
    #8A,#23,                --mov ah, [ebx]
    #43,                    --inc ebx
    #8A,#80,#00,#00,#00,#00,--mov al, [eax+lookup] (146)
    #AA,                    --stosb
    #E2,#F3,                --loop label4
    #59,                    --pop ecx
    #81,#C6,#00,#00,#00,#00,--add esi, source_add (156)
    #BA,#00,#00,#00,#00,    --mov edx, dest_add (161)
    #01,#D7,                --add edi, edx
    #01,#D3,                --add ebx, edx
    #E2,#DB,                --loop label3
    #EB,#D2})               --jmp short label2
poke4(mixcopy_proc +  5, mixcopy_proc + 146) --lookup
poke4(mixcopy_proc + 25, mixcopy_proc + 119) --height
poke4(mixcopy_proc + 55, mixcopy_proc + 136) --line
poke4(mixcopy_proc + 67, mixcopy_proc + 156) --source_add
poke4(mixcopy_proc + 79, mixcopy_proc + 161) --dest_add

constant clearmixcopy_proc = allocate_proc(
   {#8B,#58,source_mixtable,            --mov ebx, [eax+source_mixtable]
    #89,#1D,#00,#00,#00,#00,--mov [alookup], ebx (5)
    #B8}&int_to_bytes(clip_data)&{--mov eax, clip_data (1)
    #8B,#58,#0C,            --mov ebx, [eax+12]
    #83,#FB,#00,            --cmp ebx, 0
    #7D,#01,                --jge label0
    #C3,                    --ret
    #89,#1D,#00,#00,#00,#00,--label0: mov [aheight], ebx (25)
    #83,#40,#08,#04,        --add dword ptr [eax+8], 4
    #B9,#04,#00,#00,#00,    --mov ecx, 4
    #51,                    --label1: push ecx
    #56,                    --push esi
    #57,                    --push edi
    #50,                    --push eax
    #89,#C1,                --mov ecx, eax
    #FF,#49,#08,            --dec dword ptr [ecx+8]
    #8B,#59,#08,            --mov ebx, [ecx+8]
    #C1,#EB,#02,            --shr ebx, 2
    #89,#1D,#00,#00,#00,#00,--mov [aline], ebx (55)
    #8B,#06,                --mov eax, [esi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [asource_add], eax (67)
    #8B,#07,                --mov eax, [edi]
    #C1,#E8,#02,            --shr eax, 2
    #29,#D8,                --sub eax, ebx
    #A3,#00,#00,#00,#00,    --mov [adest_add], eax (79)
    #8B,#19,                --mov ebx, [ecx]
    #FF,#01,                --inc dword ptr [ecx]
    #8B,#51,#04,            --mov edx, [ecx+4]
    #FF,#56,#08,            --call near dword ptr [esi+8]
    #89,#DE,                --mov esi, ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#0C,            --call near dword ptr [edi+12]
    #53,                    --push ebx
    #8B,#59,#10,            --mov ebx, [ecx+16]
    #FF,#41,#10,            --inc dword ptr [ecx+16]
    #8B,#51,#14,            --mov edx, [ecx+20]
    #FF,#57,#08,            --call near dword ptr [edi+8]
    #5F,                    --pop edi
    #B9,#00,#00,#00,#00,    --mov ecx, height (119)
    #31,#C0,                --xor eax, eax
    #EB,#07,                --jmp label3
    #58,                    --label2: pop eax
    #5F,                    --pop edi
    #5E,                    --pop esi
    #59,                    --pop ecx
    #E2,#A1,                --loop label1
    #C3,                    --ret
    #51,                    --label3: push ecx
    #B9,#00,#00,#00,#00,    --mov ecx, line (136)
    #AC,                    --label4: lodsb
    #20,#C0,                --and al, al
    #74,#0E,                --jz label5
    #8A,#23,                --: mov ah, [ebx]
    #43,                    --inc ebx
    #8A,#80,#00,#00,#00,#00,--mov al, [eax+lookup] (150)
    #AA,                    --stosb
    #E2,#EF,                --loop label4
    #EB,#04,                --jmp short label6
    #47,                    --label5: inc edi
    #43,                    --inc ebx
    #E2,#E9,                --loop label4
    #59,                    --label6: pop ecx
    #81,#C6,#00,#00,#00,#00,--add esi, source_add (166)
    #BA,#00,#00,#00,#00,    --mov edx, dest_add (171)
    #01,#D7,                --add edi, edx
    #01,#D3,                --add ebx, edx
    #E2,#D1,                --loop label3
    #EB,#C8})               --jmp short label2
poke4(clearmixcopy_proc +  5, clearmixcopy_proc + 150) --lookup
poke4(clearmixcopy_proc + 25, clearmixcopy_proc + 119) --height
poke4(clearmixcopy_proc + 55, clearmixcopy_proc + 136) --line
poke4(clearmixcopy_proc + 67, clearmixcopy_proc + 166) --source_add
poke4(clearmixcopy_proc + 79, clearmixcopy_proc + 171) --dest_add

atom first_command, last_command
first_command = return_proc
last_command = return_proc

global procedure clear_commands()
    atom temp
    while first_command != return_proc do
        temp = peek4u(first_command + next_command)
        free(first_command)
        first_command = temp
    end while        
end procedure

global function new_command(integer kind, sequence data)
    atom handle
    handle = 0
    if kind = wait_retrace then
        handle = allocate_proc(
            command_header& 
            wait_retrace_code& 
            command_footer)
    elsif kind = copy and length(data) = 6 then
        handle = allocate(generic_length)
        poke(handle, command_header & generic_copy_code & command_footer)
        poke4(handle+generic_source_screen, current_source)
        poke4(handle+generic_dest_screen, current_dest)
        poke4(handle+generic_command_handle, handle)
        poke4(handle+generic_clip_proc, current_clip)
        poke4(handle+generic_copy_proc, copy_proc)
        poke4(handle+source_x, data)
    elsif kind = clearcopy and length(data) = 6 then
        handle = allocate(generic_length)
        poke(handle, command_header & generic_copy_code & command_footer)
        poke4(handle+generic_source_screen, current_source)
        poke4(handle+generic_dest_screen, current_dest)
        poke4(handle+generic_command_handle, handle)
        poke4(handle+generic_clip_proc, current_clip)
        poke4(handle+generic_copy_proc, clearcopy_proc)
        poke4(handle+source_x, data)
    elsif kind = destcopy and length(data) = 6 then
        handle = allocate(generic_length)
        poke(handle, command_header & generic_copy_code & command_footer)
        poke4(handle+generic_source_screen, current_source)
        poke4(handle+generic_dest_screen, current_dest)
        poke4(handle+generic_command_handle, handle)
        poke4(handle+generic_clip_proc, current_clip)
        poke4(handle+generic_copy_proc, destcopy_proc)
        poke4(handle+source_x, data)
    elsif kind = mixcopy and length(data) = 7 then
        handle = allocate(generic_mixtable_length)
        poke(handle, command_header & generic_copy_code & command_footer)
        poke4(handle+generic_source_screen, current_source)
        poke4(handle+generic_dest_screen, current_dest)
        poke4(handle+generic_command_handle, handle)
        poke4(handle+generic_clip_proc, current_clip)
        poke4(handle+generic_copy_proc, mixcopy_proc)
        poke4(handle+source_x, data)
    elsif kind = clearmixcopy and length(data) = 7 then
        handle = allocate(generic_mixtable_length)
        poke(handle, command_header & generic_copy_code & command_footer)
        poke4(handle+generic_source_screen, current_source)
        poke4(handle+generic_dest_screen, current_dest)
        poke4(handle+generic_command_handle, handle)
        poke4(handle+generic_clip_proc, current_clip)
        poke4(handle+generic_copy_proc, clearmixcopy_proc)
        poke4(handle+source_x, data)
    elsif kind = pan_to and length(data) = 2 then
        handle = allocate(source_x+8)
        poke(handle,
            command_header&
           {#8B,#1D}&int_to_bytes(handle+source_x)&--mov ebx, [source_x] (2)
           {#8B,#15}&int_to_bytes(handle+source_y)&--mov edx, [source_y] (8)
           {#FF,#15}&int_to_bytes(screen+screen_pan_to)&--call near dword ptr [pan_proc] (14)
            command_footer)
        poke4(handle+source_x, data)
    elsif kind = originate and length(data) = 2 then
        handle = allocate(source_x+8)
        poke(handle,
            command_header&
           {#8B,#1D}&int_to_bytes(handle+source_x)&--mov ebx, [source_x] (2)
           {#8B,#15}&int_to_bytes(handle+source_y)&--mov edx, [source_y] (8)
           {#FF,#15}&int_to_bytes(screen+screen_originate)&--call near dword ptr [originate_proc] (14)
            command_footer)
        poke4(handle+source_x, data)
    else
        puts(1, "Error creating command.\n")
        abort(1)
    end if
    return handle
end function

global procedure execute_commands()
    call(first_command + start_offset)
    poke(loop_command, return_proc + loop_offset)
end procedure

global procedure remove_command(atom handle)
    atom next, prev
    next = peek4u(handle + next_command)
    prev = peek4u(handle + prev_command)
    if next then
        poke4(next + prev_command, prev)
    end if
    if prev then
        poke4(prev + next_command, next)
    end if
    if handle = first_command then
        first_command = next
    end if
    if handle = last_command then
        last_command = prev
    end if
    free(handle)
end procedure

global procedure set_z_order(atom handle, integer z)
    atom next, prev
    next = peek4u(handle + next_command)
    prev = peek4u(handle + prev_command)
    if next then
        poke4(next + prev_command, prev)
    end if
    if prev then
        poke4(prev + next_command, next)
    end if
    if handle = first_command then
        first_command = next
    end if
    if handle = last_command then
        last_command = prev
    end if

    next = first_command
    while z > peek4s(next + z_order) do
        next = peek4u(next + next_command)
    end while
    
    if next = first_command then
        poke4(handle + next_command, first_command)
        poke4(handle + prev_command, 0)
        poke4(first_command + prev_command, handle)
        first_command = handle
    else
        prev = peek4u(next + prev_command)
        poke4(prev + next_command, handle)
        poke4(next + prev_command, handle)
        poke4(handle + next_command, next)
        poke4(handle + prev_command, prev)
    end if 
    poke4(handle + z_order, z)
end procedure

global procedure set_dest_pos(atom handle, sequence pos)
    poke4(handle + dest_x, pos)
end procedure

global procedure set_source_pos(atom handle, sequence pos)
    poke4(handle + source_x, pos)
end procedure

global procedure set_source_size(atom handle, sequence size)
    poke4(handle + source_w, size)
end procedure

global procedure set_mixtable(atom handle, atom mixtable)
    poke4(handle + source_mixtable, mixtable)
end procedure

global procedure set_clipping(object o)
    if atom(o) then
        if o then
            current_clip = clip_proc
        else
            current_clip = noclip_proc
        end if
    elsif length(o) = 2 then
        if o[2] then
            poke4(o[1] + generic_clip_proc, clip_proc)
        else
            poke4(o[1] + generic_clip_proc, noclip_proc)
        end if
    end if        
end procedure

global procedure select_source(object o)
    if atom(o) then
        current_source = o
    else
        poke4(o[1] + generic_source_screen, o[2])
    end if
end procedure

global procedure select_dest(atom o)
    if atom(o) then
        current_dest = o
        dest_width = peek4u(o + screen_width)
        dest_height = peek4u(o + screen_height)
    else
        poke4(o[1] + generic_dest_screen, o[2])
    end if
end procedure

-----------------------------------------------------------------
--PALETTES-------------------------------------------------------
-----------------------------------------------------------------
                    
constant set_palette_proc = allocate_proc(
  {#60,                    --pusha
   #BE,#00,#00,#00,#00,    --mov esi, pal_handle (2)
   #B0,#00,                --mov al, first_index (7)
   #BA,#C8,#03,00,#00,     --mov edx, 968
   #EE,                    --out dx, al
   #42,                    --inc edx
   #AD,                    --lodsd
   #89,#C1,                --mov ecx,eax
   #F3,#6E,                --rep outsb
   #61,                    --popa
   #C3})                   --ret
                                              
global procedure set_palette(atom handle, integer index)
    poke4(set_palette_proc + 2, handle)
    poke(set_palette_proc + 7, index)
    call(set_palette_proc)
end procedure

-- returns a handle to a palette in memory
global function get_palette(sequence pal)
    atom handle
    handle = allocate(length(pal) * 3 + 1)
    poke4(handle, length(pal) * 3)
    for i = 1 to length(pal) do
        poke(handle + i * 3 + 1, pal[i])
    end for
    return handle
end function


global function master_mix(atom source_pal, sequence source_strengths,
                           atom dest_pal,   sequence dest_strengths)
    atom handle, mixproc, squares, strengths
    handle = allocate(65536)
    squares = allocate(262144)
    for i = 0 to 255 do
        for j = 0 to 255 do
            poke4(squares+4*(i*256+j), (i-j)*(i-j))
        end for
    end for
    strengths = allocate(512)
    for i = 1 to 256 do
        poke(strengths + i*2-2, source_strengths[i])
        poke(strengths + i*2-1, dest_strengths[i])
    end for
    
    mixproc = allocate_proc(
   {#60,                    --pusha
    #BF}&int_to_bytes(handle)&{--mov edi, handle
    #BA}&int_to_bytes(dest_pal+4)&{ --mov edx, dest_pal
    #B9,#00,#01,#00,#00,    --mov ecx, 256
    #51,                    --label1: push ecx
    #B9,#00,#01,#00,#00,    --mov ecx, 256
    #BE}&int_to_bytes(source_pal+4)&{ --mov esi, source_pal
    #EB,#08,                --jmp label2
    #83,#C2,#03,            --afterlabel2: add edx, 3
    #59,                    --pop ecx
    #E2,#ED,                --loop label1
    #61,                    --popa
    #C3,                    --ret
    #51,                    --label2: push ecx
    #52,                    --push edx
    #49,                    --dec ecx
    #F6,#D9,                --neg cl
    #D1,#E1,                --shl ecx, 1
    #8B,#99}&int_to_bytes(strengths)&{--mov ebx, [ecx + strengths]
    #31,#C0,                --xor eax, eax
    #AC,                    --lodsb
    #F6,#E3,                --mul bl
    #89,#C1,                --mov ecx, eax
    #8A,#02,                --mov al,[edx]
    #F6,#E7,                --mul bh
    #01,#C8,                --add eax, ecx
    #88,#E0,                --mov al, ah
    #3C,#3F,                --cmp al, 63
    #76,#02,                --jbe skip1
    #B0,#3F,                --mov al, 63
    #50,                    --skip1: push eax
    #AC,                    --lodsb
    #F6,#E3,                --mul bl
    #89,#C1,                --mov ecx, eax
    #8A,#42,#01,            --mov al,[edx+1]
    #F6,#E7,                --mul bh
    #01,#C8,                --add eax, ecx
    #88,#E0,                --mov al, ah
    #3C,#3F,                --cmp al, 63
    #76,#02,                --jbe skip2
    #B0,#3F,                --mov al, 63
    #50,                    --skip2: push eax
    #AC,                    --lodsb
    #F6,#E3,                --mul bl
    #89,#C1,                --mov ecx, eax
    #8A,#42,#02,            --mov al,[edx+2]
    #F6,#E7,                --mul bh
    #01,#C8,                --add eax, ecx
    #88,#E0,                --mov al, ah
    #3C,#3F,                --cmp al, 63
    #76,#02,                --jbe skip3
    #B0,#3F,                --mov al, 63
    #88,#C6,                --skip3: mov dh, al
    #58,                    --pop eax
    #88,#C2,                --mov dl, al
    #58,                    --pop eax
    #BB}&int_to_bytes(dest_pal+4)&{ --mov ebx, dest_pal
    #B9,#00,#01,#00,#00,    --mov ecx, 256
    #BD,#20,#4E,#00,#00,    --mov ebp, 20000
    #EB,#07,                --jmp label3
    #5A,                    --afterlabel3: pop edx
    #59,                    --pop ecx
    #47,                    --inc edi
    #E2,#98,                --loop label2
    #EB,#8E,                --jmp afterlabel2
    #52,                    --label3: push edx
    #50,                    --push eax
    #51,                    --push ecx
    #8A,#23,                --mov ah, [ebx]
    #43,                    --inc ebx
    #C1,#E0,#02,            --shl eax, 2
    #8B,#88}&int_to_bytes(squares)&{--mov ecx, dword ptr [eax+squares]
    #88,#D0,                --mov al, dl
    #8A,#23,                --mov ah, [ebx]
    #43,                    --inc ebx
    #C1,#E0,#02,            --shl eax, 2
    #03,#88}&int_to_bytes(squares)&{--add ecx, dword ptr [eax+squares]
    #88,#F0,                --mov al, dh
    #8A,#23,                --mov ah, [ebx]
    #43,                    --inc ebx
    #C1,#E0,#02,            --shl eax, 2
    #03,#88}&int_to_bytes(squares)&{--add ecx, dword ptr [eax+squares]
    #39,#E9,                --cmp ecx, ebp
    #7F,#0D,                --jg label4
    #89,#CD,                --mov ebp, ecx
    #58,                    --pop eax
    #50,                    --push eax
    #F7,#D8,                --neg eax
    #05,#00,#01,#00,#00,    --add eax, 256
    #88,#07,                --mov [edi], al
    #59,                    --label4: pop ecx
    #58,                    --pop eax
    #5A,                    --pop edx
    #E2,#BF,                --loop label3
    #EB,#B6})               --jmp afterlabel3
    call(mixproc)
    free(mixproc)
    free(strengths)
    free(squares)
    return handle
end function                           
    
global function mix_palette(atom source_pal, atom dest_pal, integer strength)
    return master_mix(source_pal, repeat(strength, 256), dest_pal, repeat(256-strength, 256)) 
end function

global function shade_palette(sequence color, atom dest_pal)
    atom handle, source_pal
    sequence temp
    source_pal = get_palette(repeat(color, 256))
    temp = repeat(0, 256)
    for i = 1 to 256 do
        temp[i] = i-1
    end for
    handle = master_mix(source_pal, temp, dest_pal, 256-temp)
    free(source_pal)
    return handle
end function


-----------------------------------------------------------------
--PIXEL,GETPIXEL-------------------------------------------------
-----------------------------------------------------------------

global procedure pixel(object p, sequence xy)
    atom addr
    if atom(p) then
        p = {p}
    end if

----clipping is disabled for the moment----
--    if xy[2] < 0 or xy[2] >= dest_height then
--        return
--    end if
--    if xy[1] < 0 then
--        p = p[-xy[1]..length(p)]
--        xy[1] = 0
--    elsif xy[1]+length(p) >= dest_width then
--        p = p[1..dest_width - xy[1]]
--    end if
    if length(p) then
        for j = 1 to 4 do
            addr = get_dest_addr(xy)
            for i = j to length(p) by 4 do
                poke(addr, p[i])
                addr = addr + 1
            end for
            xy[1] = xy[1] + 1
        end for
    end if
end procedure

global function get_pixel(sequence xy)
    atom addr
    sequence result
    if length(xy) = 2 then
        return peek(get_source_addr(xy))
    else
        result = repeat(0, xy[3])
        for j = 1 to 4 do
            addr = get_source_addr(xy)
            for i = j to xy[3] by 4 do
                result[i] = peek(addr)
                addr = addr + 1
            end for
            xy[1] = xy[1] + 1
        end for
        return result
    end if
end function



-----------------------------------------------------------------
--INITIALIZATION-------------------------------------------------
-----------------------------------------------------------------
set_clipping(1) --on
select_source(screen)
select_dest(screen)

